@@ -1,8 +1,12 @@
 use std::borrow::Cow;
+use std::io;
 use std::sync::Arc;
-use std::time::{Duration, Instant};
-use std::{fmt, io};
+use std::time::Duration;
+#[cfg(not(target_arch = "wasm32"))]
+use std::time::Instant;
 
+#[cfg(target_arch = "wasm32")]
+use instant::Instant;
 use portable_atomic::{AtomicU64, AtomicU8, Ordering};
 
 use crate::draw_target::ProgressDrawTarget;
@@ -65,9 +69,9 @@ impl BarState {
     }
 
     pub(crate) fn reset(&mut self, now: Instant, mode: Reset) {
-        if let Reset::Eta | Reset::All = mode {
-            self.state.est.reset(now);
-        }
+        // Always reset the estimator; this is the only reset that will occur if mode is
+        // `Reset::Eta`.
+        self.state.est.reset(now);
 
         if let Reset::Elapsed | Reset::All = mode {
             self.state.started = now;
@@ -124,11 +128,12 @@ impl BarState {
     pub(crate) fn update_estimate_and_draw(&mut self, now: Instant) {
         let pos = self.state.pos.pos.load(Ordering::Relaxed);
         self.state.est.record(pos, now);
-        let _ = self.draw(false, now);
 
         for tracker in self.style.format_map.values_mut() {
             tracker.tick(&self.state, now);
         }
+
+        let _ = self.draw(false, now);
     }
 
     pub(crate) fn println(&mut self, now: Instant, msg: &str) {
@@ -139,7 +144,13 @@ impl BarState {
         };
 
         let mut draw_state = drawable.state();
-        draw_state.lines.extend(msg.lines().map(Into::into));
+        let lines: Vec<String> = msg.lines().map(Into::into).collect();
+        // Empty msg should trigger newline as we are in println
+        if lines.is_empty() {
+            draw_state.lines.push(String::new());
+        } else {
+            draw_state.lines.extend(lines);
+        }
         draw_state.orphan_lines_count = draw_state.lines.len();
         if !matches!(self.state.status, Status::DoneHidden) {
             self.style
@@ -151,6 +162,10 @@ impl BarState {
     }
 
     pub(crate) fn suspend<F: FnOnce() -> R, R>(&mut self, now: Instant, f: F) -> R {
+        if let Some((state, _)) = self.draw_target.remote() {
+            return state.write().unwrap().suspend(f, now);
+        }
+
         if let Some(drawable) = self.draw_target.drawable(true, now) {
             let _ = drawable.clear();
         }
@@ -220,13 +235,14 @@ pub struct ProgressState {
 
 impl ProgressState {
     pub(crate) fn new(len: Option<u64>, pos: Arc<AtomicPosition>) -> Self {
+        let now = Instant::now();
         Self {
             pos,
             len,
             tick: 0,
             status: Status::InProgress,
-            started: Instant::now(),
-            est: Estimator::new(Instant::now()),
+            started: now,
+            est: Estimator::new(now),
             message: TabExpandedString::NoTabs("".into()),
             prefix: TabExpandedString::NoTabs("".into()),
         }
@@ -265,8 +281,16 @@ impl ProgressState {
         };
 
         let pos = self.pos.pos.load(Ordering::Relaxed);
-        let t = self.est.seconds_per_step();
-        secs_to_duration(t * len.saturating_sub(pos) as f64)
+
+        let sps = self.est.steps_per_second(Instant::now());
+
+        // Infinite duration should only ever happen at the beginning, so in this case it's okay to
+        // just show an ETA of 0 until progress starts to occur.
+        if sps == 0.0 {
+            return Duration::new(0, 0);
+        }
+
+        secs_to_duration(len.saturating_sub(pos) as f64 / sps)
     }
 
     /// The expected total duration (that is, elapsed time + expected ETA)
@@ -274,16 +298,13 @@ impl ProgressState {
         if self.len.is_none() || self.is_finished() {
             return Duration::new(0, 0);
         }
-        self.started.elapsed() + self.eta()
+        self.started.elapsed().saturating_add(self.eta())
     }
 
     /// The number of steps per second
     pub fn per_sec(&self) -> f64 {
         if let Status::InProgress = self.status {
-            match 1.0 / self.est.seconds_per_step() {
-                per_sec if per_sec.is_nan() => 0.0,
-                per_sec => per_sec,
-            }
+            self.est.steps_per_second(Instant::now())
         } else {
             let len = self.len.unwrap_or_else(|| self.pos());
             len as f64 / self.started.elapsed().as_secs_f64()
@@ -361,80 +382,131 @@ impl TabExpandedString {
     }
 }
 
-/// Estimate the number of seconds per step
+/// Double-smoothed exponentially weighted estimator
 ///
-/// Ring buffer with constant capacity. Used by `ProgressBar`s to display `{eta}`,
-/// `{eta_precise}`, and `{*_per_sec}`.
+/// This uses an exponentially weighted *time-based* estimator, meaning that it exponentially
+/// downweights old data based on its age. The rate at which this occurs is currently a constant
+/// value of 15 seconds for 90% weighting. This means that all data older than 15 seconds has a
+/// collective weight of 0.1 in the estimate, and all data older than 30 seconds has a collective
+/// weight of 0.01, and so on.
+///
+/// The primary value exposed by `Estimator` is `steps_per_second`. This value is doubly-smoothed,
+/// meaning that is the result of using an exponentially weighted estimator (as described above) to
+/// estimate the value of another exponentially weighted estimator, which estimates the value of
+/// the raw data.
+///
+/// The purpose of this extra smoothing step is to reduce instantaneous fluctations in the estimate
+/// when large updates are received. Without this, estimates might have a large spike followed by a
+/// slow asymptotic approach to zero (until the next spike).
+#[derive(Debug)]
 pub(crate) struct Estimator {
-    steps: [f64; 16],
-    pos: u8,
-    full: bool,
-    prev: (u64, Instant),
+    smoothed_steps_per_sec: f64,
+    double_smoothed_steps_per_sec: f64,
+    prev_steps: u64,
+    prev_time: Instant,
+    start_time: Instant,
 }
 
 impl Estimator {
     fn new(now: Instant) -> Self {
         Self {
-            steps: [0.0; 16],
-            pos: 0,
-            full: false,
-            prev: (0, now),
+            smoothed_steps_per_sec: 0.0,
+            double_smoothed_steps_per_sec: 0.0,
+            prev_steps: 0,
+            prev_time: now,
+            start_time: now,
         }
     }
 
-    fn record(&mut self, new: u64, now: Instant) {
-        let delta = new.saturating_sub(self.prev.0);
-        if delta == 0 || now < self.prev.1 {
+    fn record(&mut self, new_steps: u64, now: Instant) {
+        // sanity check: don't record data if time or steps have not advanced
+        if new_steps <= self.prev_steps || now <= self.prev_time {
             // Reset on backwards seek to prevent breakage from seeking to the end for length determination
             // See https://github.com/console-rs/indicatif/issues/480
-            if new < self.prev.0 {
+            if new_steps < self.prev_steps {
+                self.prev_steps = new_steps;
                 self.reset(now);
             }
             return;
         }
 
-        let elapsed = now - self.prev.1;
-        let divisor = delta as f64;
-        let mut batch = 0.0;
-        if divisor != 0.0 {
-            batch = duration_to_secs(elapsed) / divisor;
-        };
+        let delta_steps = new_steps - self.prev_steps;
+        let delta_t = duration_to_secs(now - self.prev_time);
 
-        self.steps[self.pos as usize] = batch;
-        self.pos = (self.pos + 1) % 16;
-        if !self.full && self.pos == 0 {
-            self.full = true;
-        }
+        // the rate of steps we saw in this update
+        let new_steps_per_second = delta_steps as f64 / delta_t;
 
-        self.prev = (new, now);
+        // update the estimate: a weighted average of the old estimate and new data
+        let weight = estimator_weight(delta_t);
+        self.smoothed_steps_per_sec =
+            self.smoothed_steps_per_sec * weight + new_steps_per_second * (1.0 - weight);
+
+        // An iterative estimate like `smoothed_steps_per_sec` is supposed to be an exponentially
+        // weighted average from t=0 back to t=-inf; Since we initialize it to 0, we neglect the
+        // (non-existent) samples in the weighted average prior to the first one, so the resulting
+        // average must be normalized. We normalize the single estimate here in order to use it as
+        // a source for the double smoothed estimate. See comment on normalization in
+        // `steps_per_second` for details.
+        let delta_t_start = duration_to_secs(now - self.start_time);
+        let total_weight = 1.0 - estimator_weight(delta_t_start);
+        let normalized_smoothed_steps_per_sec = self.smoothed_steps_per_sec / total_weight;
+
+        // determine the double smoothed value (EWA smoothing of the single EWA)
+        self.double_smoothed_steps_per_sec = self.double_smoothed_steps_per_sec * weight
+            + normalized_smoothed_steps_per_sec * (1.0 - weight);
+
+        self.prev_steps = new_steps;
+        self.prev_time = now;
     }
 
+    /// Reset the state of the estimator. Once reset, estimates will not depend on any data prior
+    /// to `now`. This does not reset the stored position of the progress bar.
     pub(crate) fn reset(&mut self, now: Instant) {
-        self.pos = 0;
-        self.full = false;
-        self.prev = (0, now);
+        self.smoothed_steps_per_sec = 0.0;
+        self.double_smoothed_steps_per_sec = 0.0;
+
+        // only reset prev_time, not prev_steps
+        self.prev_time = now;
+        self.start_time = now;
     }
 
-    /// Average time per step in seconds, using rolling buffer of last 15 steps
-    fn seconds_per_step(&self) -> f64 {
-        let len = self.len();
-        self.steps[0..len].iter().sum::<f64>() / len as f64
-    }
+    /// Average time per step in seconds, using double exponential smoothing
+    fn steps_per_second(&self, now: Instant) -> f64 {
+        // Because the value stored in the Estimator is only updated when the Estimator receives an
+        // update, this value will become stuck if progress stalls. To return an accurate estimate,
+        // we determine how much time has passed since the last update, and treat this as a
+        // pseudo-update with 0 steps.
+        let delta_t = duration_to_secs(now - self.prev_time);
+        let reweight = estimator_weight(delta_t);
 
-    fn len(&self) -> usize {
-        match self.full {
-            true => 16,
-            false => self.pos as usize,
-        }
-    }
-}
+        // Normalization of estimates:
+        //
+        // The raw estimate is a single value (smoothed_steps_per_second) that is iteratively
+        // updated. At each update, the previous value of the estimate is downweighted according to
+        // its age, receiving the iterative weight W(t) = 0.1 ^ (t/15).
+        //
+        // Since W(Sum(t_n)) = Prod(W(t_n)), the total weight of a sample after a series of
+        // iterative steps is simply W(t_e) - W(t_b), where t_e is the time since the end of the
+        // sample, and t_b is the time since the beginning. The resulting estimate is therefore a
+        // weighted average with sample weights W(t_e) - W(t_b).
+        //
+        // Notice that the weighting function generates sample weights that sum to 1 only when the
+        // sample times span from t=0 to t=inf; but this is not the case. We have a first sample
+        // with finite, positive t_b = t_f. In the raw estimate, we handle times prior to t_f by
+        // setting an initial value of 0, meaning that these (non-existent) samples have no weight.
+        //
+        // Therefore, the raw estimate must be normalized by dividing it by the sum of the weights
+        // in the weighted average. This sum is just W(0) - W(t_f), where t_f is the time since the
+        // first sample, and W(0) = 1.
+        let delta_t_start = duration_to_secs(now - self.start_time);
+        let total_weight = 1.0 - estimator_weight(delta_t_start);
 
-impl fmt::Debug for Estimator {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.debug_struct("Estimate")
-            .field("steps", &&self.steps[..self.len()])
-            .field("prev", &self.prev)
-            .finish()
+        // Generate updated values for `smoothed_steps_per_sec` and `double_smoothed_steps_per_sec`
+        // (sps and dsps) without storing them. Note that we normalize sps when using it as a
+        // source to update dsps, and then normalize dsps itself before returning it.
+        let sps = self.smoothed_steps_per_sec * reweight / total_weight;
+        let dsps = self.double_smoothed_steps_per_sec * reweight + sps * (1.0 - reweight);
+        dsps / total_weight
     }
 }
 
@@ -546,6 +618,35 @@ impl Default for ProgressFinish {
     }
 }
 
+/// Get the appropriate dilution weight for Estimator data given the data's age (in seconds)
+///
+/// Whenever an update occurs, we will create a new estimate using a weight `w_i` like so:
+///
+/// ```math
+/// <new estimate> = <previous estimate> * w_i + <new data> * (1 - w_i)
+/// ```
+///
+/// In other words, the new estimate is a weighted average of the previous estimate and the new
+/// data. We want to choose weights such that for any set of samples where `t_0, t_1, ...` are
+/// the durations of the samples:
+///
+/// ```math
+/// Sum(t_i) = ews ==> Prod(w_i) = 0.1
+/// ```
+///
+/// With this constraint it is easy to show that
+///
+/// ```math
+/// w_i = 0.1 ^ (t_i / ews)
+/// ```
+///
+/// Notice that the constraint implies that estimates are independent of the durations of the
+/// samples, a very useful feature.
+fn estimator_weight(age: f64) -> f64 {
+    const EXPONENTIAL_WEIGHTING_SECONDS: f64 = 15.0;
+    0.1_f64.powf(age / EXPONENTIAL_WEIGHTING_SECONDS)
+}
+
 fn duration_to_secs(d: Duration) -> f64 {
     d.as_secs() as f64 + f64::from(d.subsec_nanos()) / 1_000_000_000f64
 }
@@ -570,31 +671,33 @@ mod tests {
     use super::*;
     use crate::ProgressBar;
 
+    // https://github.com/rust-lang/rust-clippy/issues/10281
+    #[allow(clippy::uninlined_format_args)]
     #[test]
-    fn test_time_per_step() {
+    fn test_steps_per_second() {
         let test_rate = |items_per_second| {
             let mut now = Instant::now();
             let mut est = Estimator::new(now);
             let mut pos = 0;
 
-            for _ in 0..est.steps.len() {
+            for _ in 0..20 {
                 pos += items_per_second;
                 now += Duration::from_secs(1);
                 est.record(pos, now);
             }
-            let avg_seconds_per_step = est.seconds_per_step();
+            let avg_steps_per_second = est.steps_per_second(now);
 
-            assert!(avg_seconds_per_step > 0.0);
-            assert!(avg_seconds_per_step.is_finite());
+            assert!(avg_steps_per_second > 0.0);
+            assert!(avg_steps_per_second.is_finite());
 
-            let expected_rate = 1.0 / items_per_second as f64;
-            let absolute_error = (avg_seconds_per_step - expected_rate).abs();
+            let absolute_error = (avg_steps_per_second - items_per_second as f64).abs();
+            let relative_error = absolute_error / items_per_second as f64;
             assert!(
-                absolute_error < f64::EPSILON,
-                "Expected rate: {}, actual: {}, absolute error: {}",
-                expected_rate,
-                avg_seconds_per_step,
-                absolute_error
+                relative_error < 1.0 / 1e9,
+                "Expected rate: {}, actual: {}, relative error: {}",
+                items_per_second,
+                avg_steps_per_second,
+                relative_error
             );
         };
 
@@ -610,24 +713,50 @@ mod tests {
     }
 
     #[test]
-    fn test_duration_stuff() {
-        let duration = Duration::new(42, 100_000_000);
-        let secs = duration_to_secs(duration);
-        assert_eq!(secs_to_duration(secs), duration);
+    fn test_double_exponential_ave() {
+        let mut now = Instant::now();
+        let mut est = Estimator::new(now);
+        let mut pos = 0;
+
+        // note: this is the default weight set in the Estimator
+        let weight = 15;
+
+        for _ in 0..weight {
+            pos += 1;
+            now += Duration::from_secs(1);
+            est.record(pos, now);
+        }
+        now += Duration::from_secs(weight);
+
+        // The first level EWA:
+        //   -> 90% weight @ 0 eps, 9% weight @ 1 eps, 1% weight @ 0 eps
+        //   -> then normalized by deweighting the 1% weight (before -30 seconds)
+        let single_target = 0.09 / 0.99;
+
+        // The second level EWA:
+        //   -> same logic as above, but using the first level EWA as the source
+        let double_target = (0.9 * single_target + 0.09) / 0.99;
+        assert_eq!(est.steps_per_second(now), double_target);
     }
 
     #[test]
     fn test_estimator_rewind_position() {
-        let now = Instant::now();
+        let mut now = Instant::now();
         let mut est = Estimator::new(now);
-        est.record(0, now);
-        est.record(1, now);
-        assert_eq!(est.len(), 1);
-        // Should not panic.
-        est.record(0, now);
-        // Assert that the state of the estimator reset on rewind
-        assert_eq!(est.len(), 0);
 
+        now += Duration::from_secs(1);
+        est.record(1, now);
+
+        // should not panic
+        now += Duration::from_secs(1);
+        est.record(0, now);
+
+        // check that reset occurred (estimator at 1 event per sec)
+        now += Duration::from_secs(1);
+        est.record(1, now);
+        assert_eq!(est.steps_per_second(now), 1.0);
+
+        // check that progress bar handles manual seeking
         let pb = ProgressBar::hidden();
         pb.set_length(10);
         pb.set_position(1);
@@ -636,6 +765,29 @@ mod tests {
         pb.set_position(0);
     }
 
+    #[test]
+    fn test_reset_eta() {
+        let mut now = Instant::now();
+        let mut est = Estimator::new(now);
+
+        // two per second, then reset
+        now += Duration::from_secs(1);
+        est.record(2, now);
+        est.reset(now);
+
+        // now one per second, and verify
+        now += Duration::from_secs(1);
+        est.record(3, now);
+        assert_eq!(est.steps_per_second(now), 1.0);
+    }
+
+    #[test]
+    fn test_duration_stuff() {
+        let duration = Duration::new(42, 100_000_000);
+        let secs = duration_to_secs(duration);
+        assert_eq!(secs_to_duration(secs), duration);
+    }
+
     #[test]
     fn test_atomic_position_large_time_difference() {
         let atomic_position = AtomicPosition::new();
